function [A,X] = GEMS(params)
%=============================================
% Graph Enhanced Multi-Scale Dictionary Learning algorithm (GEMS).
%
% GEMS solves the optimization problem:
%       min  |Y-Phi*A*X|_F^2 + alpha*tr(A'*Phi'*L*Phi*A) + beta*tr(X*Lc*X')  
%       A,X
%       s.t.  |x_i|_0 <= T, |a_j|_0 <= P
%
% Main parameters:
%       Phi - base dictionary over which the sparse dictionary should be represented
%       A - initial sparse dictionary (optional)
%       K - dictionary size  (if specified without the paramterer A, the initial
%                            dictionary would be sparse-coding over the base
%                            dictionary of K randomly selected training signals)
%       Y - training data
%       T - sparsity constraint (max. number of coefficients per signal)
%       P - atom sparsity constraint (max. number of coefficients per trained atom)
%       L - graph Laplacian
%       Lc - manifold graph Laplacian
%       alpha,beta - regularization coefficients
%       iternum - number of iterations (default: 10)
%
% Sparse coding parameters:
%       SCiternum - number of ADMM iterations
%       rho - ADMM step size parameter
%
% Outputs:
%       A - sparse dictionary representation matrix
%       X - sparse signal representation matrix
%
% References:
%    Y. Yankelevsky and M. Elad, "Finding GEMS: Multi-Scale Dictionaries
%    for High-Dimensional Graph Signals", IEEE Transactions on Signal
%    Processing, vol. 67, no. 7, pp. 1889-1901, April 2019.
%
% Yael Yankelevsky
% Computer Science Department
% Technion - IIT
% yaelyan@cs.technion.ac.il
% February 2019
%=============================================

if ~exist('omp','file')
    error('OMP Package missing!');
end
if ~exist('ksvds','file')
    error('KSVDS Package missing!');
end
if ~exist('GRSC_ADMM','file')
    error('DGRDL Package missing!');
end

global memusage
if isfield(params,'memusage')
    memusage = params.memusage;
else
    memusage = 'normal';
end

if isfield(params,'Y')
    Y = params.Y;
else
    error('Input data matrix Y missing!');
end
[N,M] = size(Y); % N = signal dimension, M = number of signals

if isfield(params,'Phi')
    Phi = params.Phi;
else
    error('Input base dictionary Phi missing!');
end
% verify base dictionary normalization
if (any(abs(sum(Phi.*Phi)-1) > 1e-2))
    error('Base dictionary atoms are not normalized to unit length');
end
if (size(Phi,1) ~= N)
    error('Base dictionary size is incompatible with signal size.');
end
baseG = Phi'*Phi;

if isfield(params,'T')
    T = params.T;
else
    error('Sparsity constraint T missing!');
end
if isfield(params,'P')
    Tdict = params.P;
else
    error('Atom sparsity constraint P missing!');
end

% dictionary initialization
if (isfield(params,'A'))
    A = params.A;
    K = size(A,2); % number of atoms
    if (isfield(params,'K'))
        if (params.K ~= K)
            error('Invalid initial matrix A: size does not match the field K');
        end
    end
    if (size(A,1) ~= size(Phi,2))
        error('Invalid initial matrix A');
    end
elseif (isfield(params,'K'))
        K = params.K;
        data_ids = find(colnorms_squared(Y)>1e-6);   % ensure no zero data elements are chosen
        perm = randperm(length(data_ids));
        A = omp(Phi,Y(:,data_ids(perm(1:K))),baseG,Tdict,'checkdict','off');
else
    error('Missing input dictionary or dictionary size!');
end
A = normalizeAcols(Phi,A); % make sure the atoms are normalized
if (M < K)
    error('Number of training signals is smaller than number of atoms to train');
end

if isfield(params,'alpha')
    alpha = params.alpha;
elseif isfield(params,'L')
    error('Regularizaion coefficient alpha missing!');
else
    alpha = 0;
end

if isfield(params,'beta')
    beta = params.beta;
elseif isfield(params,'Lc')
    error('Regularizaion coefficient beta missing!');
else
    beta = 0;
end

if isfield(params,'L')
    L = params.L;
elseif alpha~=0
    error('Graph Laplacian L missing!');
end

if isfield(params,'Lc')
    Lc = params.Lc;
elseif beta~=0
    error('Manifold Laplacian Lc missing!');
end

if isfield(params,'iternum')
    iternum = params.iternum;
else
    iternum = 10;
end

if isfield(params,'sdebug') % activate debug mode
    sdebug = params.sdebug;
else
    sdebug = 0;
end

% for efficiency, run ksvds if alpha=beta=0
if alpha == 0 && beta == 0
    ksvds_params = params;
    ksvds_params.data = Y;
    if isfield(params,'A')
        ksvds_params.initA = params.A;
    end
    B = {};
    B{1} = Phi;
    B{2} = [1];
    ksvds_params.basedict = B;
    ksvds_params.Tdata = T;
    ksvds_params.Tdict = Tdict;
    ksvds_params.dictsize = K;
    [A,X] = ksvds(ksvds_params, '');
    return
end

% sparse coding parameters
scparams.D = Phi*A;
scparams.Y = Y;
scparams.T = T;
scparams.L = Lc;
scparams.beta = beta;
if isfield(params,'SCiternum')
    scparams.iternum = params.SCiternum;
else
    scparams.iternum = 20; % default
end
if isfield(params,'rho')
    scparams.rho = params.rho;
else
    scparams.rho = 1e3;
end
scparams.debug = sdebug;
scparams.runDebiasing = 1; % run debiasing of sparse representations

% atom update parameters
atom_params = scparams;
atom_params.runDebiasing = 0; % no debiasing for atom update
atom_params.normalize = 1; % normalize atoms

muthresh = 0.99;
ompparams = {'checkdict','off'};
err = zeros(1,iternum);
for iter=1:iternum
    G = A'*baseG*A;
    % sparse coding
    if (beta==0)
        X = omp(Phi*A,Y,G,T,ompparams{:});
    else
        X = GRSC_ADMM(scparams);
    end
    % dictionary update
    replaced_atoms = zeros(1,K);  % track replaced atoms
    unused_sigs = 1:M;  % tracks the signals used for replacing atoms
    atomperm = randperm(K);
    for j=1:K
        [A(:,atomperm(j)),x_j,data_indices,unused_sigs,replaced_atoms] = optimize_atom(Y,Phi,A,atomperm(j),X,baseG,Tdict,unused_sigs,replaced_atoms,L,Lc,alpha,beta,atom_params);
        X(atomperm(j),data_indices) = x_j; % plug back into X
    end
    err(iter) = compute_err(Phi*A,X,Y);
    
    [A,cleared_atoms] = cleardict(Phi,A,X,Y,baseG,Tdict,muthresh,unused_sigs,replaced_atoms,L,alpha,atom_params);

    fprintf('Iteration %d / %d complete, RMSE = %.4g\n', iter, iternum, err(iter));
    
    totalReplaced = sum(replaced_atoms) + cleared_atoms;
    if (totalReplaced>0 && sdebug)
        fprintf('   replaced %d atoms\n', totalReplaced);
    end
    
    scparams.D = Phi*A;
end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           Optimize atom              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [atom,x_j,data_indices,unused_sigs,replaced_atoms] = optimize_atom(Y,Phi,A,j,X,baseG,Tdict,unused_sigs,replaced_atoms,L,Lc,alpha,beta,params)
[x_j,data_indices] = sprow(X,j);
smallX = X(:,data_indices);
Aj = A(:,j);

if (length(data_indices) < 1) %found unused atom
    maxsignals = 5000;
    perm = randperm(length(unused_sigs));
    perm = perm(1:min(maxsignals,end));
    E = sum((Y(:,unused_sigs(perm)) - Phi*A*X(:,unused_sigs(perm))).^2);
    [~,i] = max(E);
    if (alpha==0)
        atom = omp(Phi,Y(:,unused_sigs(perm(i))),baseG,Tdict,'checkdict','off');
    else
        atom = update_atom_L(Y(:,unused_sigs(perm(i))),Phi,baseG,L,Tdict,alpha,params);
    end
    d = Phi*atom;
    atom = atom./norm(d);
    x_j = zeros(size(x_j));
    unused_sigs = unused_sigs([1:perm(i)-1,perm(i)+1:end]);
    replaced_atoms(j) = 1;
    return
end

EjR = (Y(:,data_indices)-Phi*A*smallX+Phi*Aj*x_j)*x_j';
% update atom j
if (alpha==0)
    atom = omp(Phi,EjR,baseG,Tdict,'checkdict','off');
else
    EjR = EjR/norm(x_j); % x_j should be normalized
    atom = update_atom_L(EjR,Phi,baseG,L,Tdict,alpha,params);
end
d = Phi*atom;
atom = atom./norm(d);
d = d./norm(d);

% update x_j = g^T = E_j^T(R)*Phi*a_j
x_j = d'*Y(:,data_indices) - d'*Phi*A*smallX + Aj'*Phi'*d*x_j;

if (beta~=0)
    LcR = Lc(data_indices,data_indices);
    x_j = (eye(size(LcR))+beta*LcR)\x_j';
    x_j = x_j'; % make it a row vector again
end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%             Atom update              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function a = update_atom_L(E,B,G,L,P,alpha,params)
% Task:
% min_a |E-Ba|_2^2+alpha*a'Ma s.t. |a|_0<=P
% Inputs:
% E - Ej*g=Ej*xj [Nx1] (xj=j-th row of X rotated to column)
% B - Phi [NxK]
% G - Phi'*Phi [KxK]
% L - [NxN]==> M = Phi'*L*Phi [KxK]
% P - atom sparsity threshold
% alpha - regularization coefficient
global memusage
if strcmp(memusage,'high') % use ADMM
    a = atom_pursuit_ADMM(E,B,G,B'*L*B,P,alpha,params);
else % use modified greedy pursuit
    a = orthoRegOMP(E,B'*L*B,alpha,P);
end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           cleardict                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [A,cleared_atoms] = cleardict(Phi,A,X,Y,baseG,Tdict,muthresh,unused_sigs,replaced_atoms,L,alpha,params)
use_thresh = 4;  % at least this number of samples must use the atom to be kept
K = size(A,2);

% compute error in blocks to conserve memory
err = zeros(1,size(Y,2));
blocks = [1:3000:size(Y,2) size(Y,2)+1];
for i = 1:length(blocks)-1
    err(blocks(i):blocks(i+1)-1) = sum((Y(:,blocks(i):blocks(i+1)-1)-Phi*A*X(:,blocks(i):blocks(i+1)-1)).^2);
end

cleared_atoms = 0;
usecount = sum(abs(X)>1e-7, 2);
for j = 1:K
    % compute G(:,j)
    if (numel(baseG)>0)
        Gj = A'*(baseG*A(:,j));
    else
        Gj = A'*(Phi'*Phi)*A(:,j);
    end
    Gj(j) = 0;
    % replace atom
    if ( (max(Gj.^2)>muthresh^2 || usecount(j)<use_thresh) && ~replaced_atoms(j) )
        [~,i] = max(err(unused_sigs));
        if (alpha==0)
            atom = omp(Phi,Y(:,unused_sigs(i)),baseG,Tdict,'checkdict','off');
        else
            atom = update_atom_L(Y(:,unused_sigs(i)),Phi,baseG,L,Tdict,alpha,params);
        end
        d = Phi*atom;
        atom = atom./norm(d);
        A(:,j) = atom;

        unused_sigs = unused_sigs([1:i-1,i+1:end]);
        cleared_atoms = cleared_atoms+1;
    end
end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%             compute_err              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function err = compute_err(D,X,Y)
err2 = zeros(1,size(Y,2));
blocksize = 2000;
for i = 1:blocksize:size(Y,2)
    blockids = i : min(i+blocksize-1,size(Y,2));
    err2(blockids) = sum((Y(:,blockids) - D*X(:,blockids)).^2);
end
err = sqrt(sum(err2)/numel(Y));
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            misc functions            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Y = colnorms_squared(X)
Y = zeros(1,size(X,2));
blocksize = 2000;
for i = 1:blocksize:size(X,2)
    blockids = i : min(i+blocksize-1,size(X,2));
    Y(blockids) = sum(X(:,blockids).^2);
end
end

function A = normalizeAcols(Phi,A)
atomnorms = zeros(size(A,2),1);
for i = 1:size(A,2)
    atomnorms(i) = sqrt(sum((Phi*A(:,i)).^2));
end
A = A*spdiags(1./atomnorms,0,size(A,2),size(A,2));
end
