function [A,X,L,Phi] = GEMS_optL(params)
%=============================================
% Graph Enhanced Multi-Scale Dictionary Learning algorithm (GEMS)
% with optimized Laplacian.
%
% GEMS_optL solves the optimization problem:
%       min  |Y-Phi*A*X|_F^2 + alpha*tr(A'*Phi'*L*Phi*A) + beta*tr(X*Lc*X')+ mu*|L|_F^2  
%      A,X,L
%       s.t.  |x_i|_0 <= T
%             |a_j|_0 <= P
%             L(i,j)=L(j,i)<=0 (i~=j)
%             L*ones(N,1) = zeros(N,1)
%             trace(L) = N
%
% GEMS parameters:
%       Phi - initial base dictionary 
%       A - initial sparse dictionary (optional)
%       K - dictionary size  (if specified without the paramterer A, the initial
%                            dictionary would be sparse-coding over the base
%                            dictionary of K randomly selected training signals)
%       Y - training data
%       T - sparsity constraint (max. number of coefficients per signal)
%       P - atom sparsity constraint (max. number of coefficients per trained atom)
%       L - graph Laplacian
%       Lc - manifold graph Laplacian
%       alpha,beta - regularization coefficients
%       iternum - number of iterations (default: 10)
%
% Sparse coding parameters:
%       SCiternum - number of ADMM iterations
%       rho - ADMM step size parameter
%
% Graph learning parameters:
%       mu - regularization coefficient
%       Liter - frequency of Laplacian updates (default: 4)
%       thr - weight threshold (discard edges with weight<thr)
%
% Outputs:
%       A - sparse dictionary representation matrix
%       X - sparse signal representation matrix
%       L - learned graph Laplacian
%       Phi - learned graph-wavelet basis
%
% References:
%    Y. Yankelevsky and M. Elad, "Finding GEMS: Multi-Scale Dictionaries
%    for High-Dimensional Graph Signals", IEEE Transactions on Signal
%    Processing, vol. 67, no. 7, pp. 1889-1901, April 2019.
%
% Yael Yankelevsky
% Computer Science Department
% Technion - IIT
% yaelyan@cs.technion.ac.il
% February 2019
%=============================================

if isfield(params,'L')
    L = params.L;
else
    error('Initial graph Laplacian L missing!');
end

if isfield(params,'Phi')
    Phi = params.Phi;
else
    error('Base dictionary Phi missing!');
end

if isfield(params,'alpha')
    alpha = params.alpha;
else
    error('Regularizaion coefficient alpha missing!');
end

if alpha == 0
    alphaL = 1;
else
    alphaL = alpha;
end
if isfield(params,'mu')
    mu = params.mu;
else
    mu = 0.8*alphaL;
end

if isfield(params,'iternum')
    iternum = params.iternum;
else
    iternum = 20;
end

if isfield(params,'Liter')
    Liter = params.Liter;
else
    Liter = 4;
end

innerIter = round(iternum/Liter);
params.iternum = innerIter;
Lparams.alpha = alphaL;
Lparams.mu = mu;
Lparams.L = L;
if isfield(params,'thr')
    Lparams.thr = params.thr;
end
for iter=1:Liter
    fprintf('Global update iteration %d/%d\n',iter,Liter);
    [A,X] = GEMS(params);
    if (iter<Liter)  % update L and Phi for the following iterations
        curD = Phi*A;
        Lparams.D = curD;
        L = optimizeGraphLaplacian(Lparams);
        Lparams.L = L;
        W = diag(diag(L))-L;
        Phi = graphHaarWaveletConstruction(W);
        params.Phi = Phi;
        params.A = A;
        params.L = L;
    end
end
end
