function a = atom_pursuit_ADMM(E,B,G,M,P,alpha,params)
%=============================================
% Sparse atom update using ADMM
%
% atom_pursuit_ADMM solves the optimization problem:
% 	min |E-B*a|_2^2+alpha*a'*M*a  s.t. |a|_0 <= P
%    a
%
% Inputs:
% E - Ej*g=Ej*xj [Nx1] (xj=j-th row of X rotated to column)
% B - Phi [NxK]
% G - Phi'*Phi [KxK]
% M - Phi'*L*Phi [KxK] (L is [NxN])
% P - atom sparsity threshold
% alpha - regularization coefficient
% params - struct of optional parameters:
%       iternum - number of ADMM iterations (default = 20)
%       rho - ADMM parameter (default=1000)
%       debug - print debug messages (default = 0)
%       runDebiasing - fix values on the support using least-squares (default = 0)
%       normalize - return a normalized vector (default = 1)
%
% References:
%    Y. Yankelevsky and M. Elad, "Finding GEMS: Multi-Scale Dictionaries
%    for High-Dimensional Graph Signals", IEEE Transactions on Signal
%    Processing, vol. 67, no. 7, pp. 1889-1901, April 2019.
%
% Yael Yankelevsky
% Computer Science Department
% Technion - IIT
% yaelyan@cs.technion.ac.il
% February 2019
%=============================================

if isfield(params,'iternum')
    iternum = params.iternum;
else
    iternum = 20;
end
if isfield(params,'rho')
    rho = params.rho;
else
    rho = 1e3;
end
if isfield(params,'debug')
    debugMode = params.debug;
else
    debugMode = 0;
end
if isfield(params,'runDebiasing')
    runDebiasing = params.runDebiasing;
else
    runDebiasing = 0;
end
if isfield(params,'normalize')
    doNormalize = params.normalize;
else
    doNormalize = 1;
end


% initialize
K = size(G,1);
a = zeros(K,1);
b = a;
u = zeros(K,1);
objValue = zeros(iternum,1);
for i = 1:iternum
    % update a
    a = (G+alpha*M+rho*eye(K))\(B'*E+rho*(b-u));
    % update b
    b = zeros(K,1);
    AU = a+u;
    [~,ind] = sort(abs(AU),'descend');
    b(ind(1:P)) = AU(ind(1:P));
    % update u
    u = u+a-b;
    if (debugMode>0)
        objValue(i) = norm(E-B*b,'fro')^2+alpha*a'*M*a;
        fprintf('Iter %d objective=%.4f\n',i,objValue(i));
    end
end
if (runDebiasing)
    suppInd = find(b~=0);
    y = B'*E;
    ys = y(suppInd);
    Ms = M(suppInd,suppInd);
    xd = zeros(size(b));
    xd(suppInd) = (eye(size(Ms))+alpha*Ms)\ys;
    b = xd;
end
if (doNormalize)
    b = b/norm(B*b); % make sure the final atoms are normalized
end
a = sparse(b); % return sparse outcome
end
