function Phi = graphHaarWaveletConstruction(W)
%=============================================
% Recursive construction of a graph-Haar wavelet basis
%
% Input arguments:
%       W - weight matrix of the graph
% Output arguments:
%       Phi - graph-Haar wavelet basis
%
% References:
%    Y. Yankelevsky and M. Elad, "Finding GEMS: Multi-Scale Dictionaries
%    for High-Dimensional Graph Signals", IEEE Transactions on Signal
%    Processing, vol. 67, no. 7, pp. 1889-1901, April 2019.
%
% Yael Yankelevsky
% Computer Science Department
% Technion - IIT
% yaelyan@cs.technion.ac.il
% February 2019
%=============================================

N = size(W,1);
H = ones(N,1)/sqrt(N); 
Phi = treeHaarRecurse(H,W,1:N,N);
end


function H = treeHaarRecurse(H,W,wind,N)
% H is the current graph-Haar matrix
% W is the current weight matrix of the graph
% wind are the node indices
% N is the total number of nodes (in the full graph)

fvec = estimateFiedlerVec(W);

thr = 1e-15;
posind = find(fvec>thr);
negind = find(fvec<=thr);
count(1) = length(posind);
count(2) = length(negind); %length(x)-count(1);
normc = count/sum(count);
if (~all(count>0))
    warning('vector could not be partitioned\n');
    fprintf(2,'WARNING, vector could not be partitioned\n');
    return
end
newH = zeros(N,1);
newH(wind(posind)) = 1*normc(2);
newH(wind(negind)) = -1*normc(1);
H = [H,newH/norm(newH)];
if count(1)>1
    H = treeHaarRecurse(H,W(posind,posind),wind(posind),N);
end
if count(2)>1
    H = treeHaarRecurse(H,W(negind,negind),wind(negind),N);
end
end


function fvec = estimateFiedlerVec(W)
L = diag(sum(W,2))-W;
N = size(L,1);
if (N>1000) % approximate only this eigenvector
    [V,E] = eigs(L,2,'sm');
    fvec = V(:,1);
else % run the more-accurate full eigendecomposition
     % (eigs has some numerical squirks)
    [V,~] = eig(L);
    fvec = V(:,2);
end
end