function x = orthoRegOMP(y,M,alpha,T0,runDebiasing)
%=============================================
% Sparse atom update using a modified greedy pursuit
%
% orthoRegOMP solves the optimization problem:
% x = argmin ||y-x||_2^2+alpha*x'*M*x   s.t. ||x||_0=T0
%
% Input arguments:
%       y - input vector to represent
%       M - effective Laplacian (Phi'*L*Phi)
%       alpha - regularization coefficient
%       T0 - maximal number of non-zero coefficients (sparsity threshold)
% Output arguments:
%       x - sparse coefficient vector
%
% References:
%    Y. Yankelevsky and M. Elad, "Finding GEMS: Multi-Scale Dictionaries
%    for High-Dimensional Graph Signals", IEEE Transactions on Signal
%    Processing, vol. 67, no. 7, pp. 1889-1901, April 2019.
%
% Yael Yankelevsky
% Computer Science Department
% Technion - IIT
% yaelyan@cs.technion.ac.il
% February 2019
%=============================================

if (nargin<5)
    runDebiasing = 0;
end
N = length(y);
x = zeros(N,1);
residual = y;
indx = zeros(T0,1);
% compute the sparse representation using greedy pursuit
for k=1:T0
    proj = zeros(N,1);
    temp1 = zeros(N,1);
    temp2 = zeros(N,1);
    for j=1:N
        if find(indx==j)
            continue;
        end
        temp1(j) = residual(j)-alpha*x'*M(:,j);
        temp2(j) = 1+alpha*M(j,j);
        proj(j) = (temp1(j))^2/temp2(j);
    end
    assert(all(proj>=0))
    [~,pos] = max(abs(proj));
    indx(k) = pos;
    x(pos) = temp1(pos)/temp2(pos);
    residual = y-x;
%     if sum(residual.^2) < 1e-6
%         break;
%     end
end
if (runDebiasing)
    suppInd = find(x~=0);
    ys = y(suppInd);
    Ms = M(suppInd,suppInd);
    xd = zeros(N,1);
    xd(suppInd) = (eye(size(Ms))+alpha*Ms)\ys;
    x = sparse(xd);
    return
end
x = sparse(x);
end

